/**
 * Implement the server of Online Jam Space, acting a central part between the CGI file and the database.
 * It takes arguments from the CGI file and, based on the arguments, accesses the database.
 * Simply it performs three tasks: 
 *		inserting new user to the database, 
 *		updating the online-ness, nickname, and instrument of the user, 
 *		and removing the user from the database upon his/her informed disconnection.
 * The uninformed disconnection cleaning up is performed by a comparison 
 * with the current and lastModified timestamp every time after the any of the above three actions is performed.
 * @author Kinson Liu
 * @version v2.0
 */

import java.sql.*;
import java.net.*;
import java.io.*;

public class Server{
	private final int IP = 0, PORT = 1, NICKNAME = 2, INSTRUMENT = 3, LASTMODIFIED = 5;
	private final int INSERT = 0, UPDATE = 1, DELETE = 2;
	
	/* define the possible incoming arguments */
	/* argv defines all the possible incoming arguments and the order they have to be in */
	/* insert defines the order and what arguments must be provided */
	/* same as update and delete */
	private final String argv[] = {"action", "IP", "Port", "nickname", "instrument"};
	private final String insert[] = {"IP", "Port", "nickname", "instrument"};
	private final String update[] = {"IP", "Port", "nickname", "instrument"};
	private final String delete[] = {"IP", "Port"};
	
	/* possible actions */
	private final String actions[] = {"insert", "update", "delete"};
	
	private int action = -1; // inititalized to -1, meaning undefined
	private String query[];  // storing the arguments and values according to the action
	
	/* database info */
	private String DB_Driver = "jdbc:mysql";
	private String DB_Server = "abstract";
	private String DB_DB = "yanyin_JamSpace";
	private String DB_Table = "OnlineList";
	private String DB_URL = DB_Driver+"://"+DB_Server+"/"+DB_DB;
	private String DB_User = "yanyin";
	private String DB_Pass = "71eml8q9";
	
	Connection con;
	Statement s;
	
	/* in terms of second */
	/* timeInterval means the amount of time that the client has to update its 
	/* timestamp in the database. */
	/* timeAllowance means the client can have up to that amount of time for "buffering" */
	/* that helps when there is delay during the propagation and doesn't meet the timeInterval */
	private long timeInterval;
	private long timeAllowance;
	
	/**
	 * Upon its instantiation, process the incoming string query, validate the database settings,
	 * send the timeInterval and timeAllowance, take actions (insert, update, delete) accordingly, 
	 * and send the online user list back to the CGI in XML format.
	 * @param query - the incoming query string
	 * @throws IllegalArgumentException if the incoming query string is not valid
	 * @post every kind error is handled, user list is return correctly
	 */
	public Server(String query){
		try{
			assignQuery(query);
			checkDB();
			timeInterval = 5;
			timeAllowance = 1;
			execution();
			outputList();
		}
		catch(SQLException e){			
			System.out.println("<ERROR>"+"Database: "+e.getMessage()+"("+e.getErrorCode()+")"+"</ERROR>");
		}
		catch(IllegalArgumentException e){
			System.out.println("<ERROR>"+"Query string: "+e.getMessage()+"</ERROR>");
		}
		catch(Exception e){
			System.out.println("<ERROR>"+"Others: "+e.getMessage()+"</ERROR>");
		}
	}
	
	/**
	 * Break down the incoming query string into respective arguments to the database.
	 * Argument values are stored in the query variable.
	 * The possible arguments are stored inside the argv array
	 * The possible values of action are defined in the acitons array
	 * @param query - the incoming query string
	 * @throws IllegalArgumenException - if the format or the values of the incoming query string are not valid 
	 * @pre the incoming query string is valid
	 * @post the incoming query string is validated and arguments and values are stored into query[] and action.
	 * @test any violating the following test will result an IllegalArgumentException
	 * 			1. check the minimum and maximum number of arguments
	 * 			2. check if action is defined
	 * 			3. check if the action is valid
	 * 			4. check if every pair has a key and value
	 * 			5. check if the number of arguments of that action is correct
	 * 			6. check if arguments are valid
	 */
	private void assignQuery(String query) throws IllegalArgumentException, UnsupportedEncodingException{
		String temp[] = query.split("&");
		if(temp.length<1 || temp.length>argv.length) incomingQueryError("temp.length<1 || temp.length>argv.length");	
		String temp1[] = temp[0].split("=");	
		if(!temp1[0].equals("action")) incomingQueryError("equals(\"action\")");	
		for(int i = 0; i<actions.length; i++) if(temp1[1].equals(actions[i])) action = i;
		if(action==-1) incomingQueryError("action==-1");
		String temp2[][] = new String[temp.length-1][2];
		for(int i = 0; i <temp2.length; i++){
			String temp3[] = temp[i+1].split("=");	
			if(temp3.length!=2) incomingQueryError("temp3.length!=2");		
			temp2[i][0] = temp3[0];
			temp2[i][1] = temp3[1];
		}
		switch(action){
		case INSERT:
			checkInsert(temp2);
			break;
		case UPDATE:
			checkUpdate(temp2);
			break;
		case DELETE:
			checkDelete(temp2);
		}
		this.query = new String[temp2.length];
		for(int i = 0; i<this.query.length; i++) this.query[i]=temp2[i][1];
		if(action == INSERT || action == UPDATE){
			this.query[NICKNAME] = URLDecoder.decode(this.query[NICKNAME], "UTF-8");
			this.query[INSTRUMENT] = URLDecoder.decode(this.query[INSTRUMENT], "UTF-8");
		}
	}
	
	/**
	 * check the arguments of insert
	 * @param temp2
	 * @throws IllegalArgumentException
	 */
	private void checkInsert(String temp2[][]) throws IllegalArgumentException{
		if(temp2.length!=insert.length) incomingQueryError("temp2.length!=insert.length");
		for(int i = 0; i<temp2.length; i++) if(!temp2[i][0].equals(insert[i])) incomingQueryError("!temp2[i][0].equals(insert[i])");
	}
	
	/**
	 * check the arguments of update
	 * @param temp2
	 * @throws IllegalArgumentException
	 */
	private void checkUpdate(String temp2[][]) throws IllegalArgumentException{
		if(temp2.length!=update.length) incomingQueryError("temp2.length!=update.length");
		for(int i = 0; i<temp2.length; i++) if(!temp2[i][0].equals(update[i])) incomingQueryError("!temp2[i][0].equals(update[i])");
	}
	
	/**
	 * check the arguments of delete
	 * @param temp2
	 * @throws IllegalArgumentException
	 */
	private void checkDelete(String temp2[][]) throws IllegalArgumentException{
		if(temp2.length!=delete.length) incomingQueryError("temp2.length!=delete.length");
		for(int i = 0; i<temp2.length; i++) if(!temp2[i][0].equals(delete[i])) incomingQueryError("!temp2[i][0].equals(update[i])");
	}
	
	/**
	 * Throws IllegalArgumentExceptions if there is error in the incoming query string
	 * @param e the error message
	 */
	private void incomingQueryError(String e) throws IllegalArgumentException{
		throw new IllegalArgumentException(e);
	}	
	
	/**
	 * Check the connectability of the database
	 * @throws SQLException is any error occurs
	 */
	private void checkDB() throws Exception{
		Class.forName("com.mysql.jdbc.Driver").newInstance();
		con = DriverManager.getConnection(DB_URL, DB_User, DB_Pass);
		con.close();
	}
	
	/**
	 * Based on the action to perform the action (insert, update, delete)
	 * Tell contained actions whether a user is already inside the database
	 * @pre action is one of the valid actions
	 */
	private void execution() throws Exception{	
		Connection con = DriverManager.getConnection(DB_URL, DB_User, DB_Pass);
		Statement s = con.createStatement();
		ResultSet r = s.executeQuery("select * from "+DB_Table+" where IP = \""+query[IP]+"\"and Port = "+query[PORT]+";");
		switch(action){
		case INSERT:
			insert(s, r.next());
			break;
		case UPDATE:
			update(s, r.next());
			break;
		case DELETE:
			delete(s, r.next());
		}
		s.executeUpdate("delete from "+DB_Table+" where unix_timestamp(lastModified) + "+(timeInterval+timeAllowance)+" <= unix_timestamp(now());");
		r.close();
		s.close();
		con.close();
	}
	
	/**
	 * Insert the new user (IP, Port, nickname, instrument) in the database
	 * First check whether the user is already defined. 
	 * If yes, update the lastModified
	 * If no, insert the user
	 * @throws SQLExceptoin upon any error in the database
	 */
	private void insert(Statement s, boolean userFound) throws SQLException{
		if(userFound) update(s, true);
		else{
			String attributes = null;
			for(int i = 0; i<insert.length; i++)
				if(attributes == null) attributes = insert[i];
				else attributes = attributes + ", " +insert[i];
			s.executeUpdate("insert into "+DB_Table+"("+attributes+")  values (\""+query[IP]+"\", "+query[PORT]+", \""+query[NICKNAME]+"\", \""+query[INSTRUMENT]+"\");");
		}
	}
	
	/**
	 * Update the nickname, instrument, and lastModified of the user based on the IP and Port
	 * First check whether the user is already defined. 
	 * If yes, update the lastModified
	 * If no, throw exception
	 * @throws SQLExceptoin upon any error in the database
	 */
	private void update(Statement s, boolean userFound) throws SQLException{
		if(userFound) s.executeUpdate("update "+DB_Table+" set nickname = \""+query[NICKNAME]+"\", instrument = \""+query[INSTRUMENT]+"\", lastModified = now() where IP = \""+query[IP]+"\" and Port = "+query[PORT]+";");
		else throw new SQLException("User not found, couldn't update");
	}
	
	/**
	 * delete the user based on the IP and Port	 
	 * First check whether the user is already defined. 
	 * If yes, delete the user
	 * If no, throw exception
	 * @throws SQLExceptoin upon any error in the database
	 */
	private void delete(Statement s, boolean userFound) throws SQLException{
		if(userFound) s.executeUpdate("delete from "+DB_Table+" where IP = \""+query[IP]+"\" and Port = "+query[PORT]+";");
		else throw new SQLException("User not found, couldn't delete");
	}
	
	
	/**
	 * Output the list of online user
	 */
	private void outputList() throws Exception{
		Connection con = DriverManager.getConnection(DB_URL, DB_User, DB_Pass);
		Statement s = con.createStatement();
		ResultSet r = s.executeQuery("select * from "+DB_Table+";");
		System.out.println("<UserList timeInterval=\""+timeInterval+"\" timeAllowance=\""+timeAllowance+"\">");
		while(r.next()) {
			System.out.println("\t<Response>");
			for(int i = 1; i<argv.length; i++) System.out.println("\t\t<"+argv[i]+">"+r.getString(i)+"</"+argv[i]+">");
			System.out.println("\t\t<lastModified>"+r.getString(LASTMODIFIED)+"</lastModified>");
			System.out.println("\t</Response>");
		}
		System.out.println("</UserList>");
		r.close();
		s.close();
		con.close();	 
	}
	
	public static void main(String argv[]){
		try{
			if(argv.length!=1) throw new IllegalArgumentException("No parameter list");
			new Server(argv[0]);
		}
		catch(Exception e){
			System.out.println("<ERROR>"+"Others: "+e.getMessage()+"</ERROR>");
		}
	}
}
